/* N-body acceleration calculation using Greenspan 1st-order discrete	*/
/* mechanics for comparison with heirarchical methods.		.	*/
/* 1-D version with vector calculation, mixed spreading and systolic	*/
/* computation.  All processors store particle information and		*/
/* calculate velocities, positions redundantly.				*/
/* Uses dynamic shape allocation to permit more flexible operation.	*/
/* File:  nbig.cs		Prototypes:  nbig.hs			*/
/* Header for associated files:  nbigsys.hs				*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>
#include <float.h>
#include "gbh.hs"

/* <float.h> provides useful parameters about floating-point	*/
/*  representations						*/

/* Geometry is [NM][NM][NS] with NS = systolic pipe length (even),	*/
/* NB = NM*NS,   NM*NM*NS >= physical size.				*/

int NM =  32;	/* Number of pipes in each dimension		*/
int NS =  32;	/* Number of entries in systolic pipe >= 4	*/
int SS =  16;	/* Number of systolic steps = NS/2		*/

extern int NB;	/* Number of bodies in system			*/
extern int PX;	/* Controls how much state is printed		*/
extern float G;	/* Gravitational constant			*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  All processors have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [][][]pair;	/* Will be [NM][NM][NS]	*/

#include "green.hs"
#include "compare.hs"
#include "acbig.hs"

/* All parallel variables are declared as pointers so 'pair' can be dynamic. */
/* State variables will be allocated as x0[3],...;  w, m, mj, are floats     */

/* x0, x1 - current, new positions;	v0, v1 - current, new velocities */
/* x0j - x0 for initial "j" particle					 */
/* m - mass;		mi,mj - scaled mass;	ac - acceleration	 */
/* w, wv, ax - working							 */
/* xm0 - conveyor (moving) positions; 3 entries	       \		 */
/* xm1 - conveyor (moving) positions; 3 entries		\ vector of	 */
/* acm - conveyor (moving) accelerations; 3 entries	/ 10 entries	 */
/* mmj - conveyor (moving) scaled masses; 1 entry      /		 */

float:pair *x0, *x1, *v0, *v1, *m, *mi, *mj, *ac, *ax, *w, *wv,
	   *x0j, *xm0, *xm1, *acm, *mmj;

/* i - row number;    j - column number;    k - pipe position	*/

int:pair *i, *j, *k;

/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/

bool:pair *diag, *act;

/* Data sizes for allocation, communication */

int fsz, isz, bsz, vsz, cvsz;


/* These are not used, but introduced for compatibility */

float:pair  *xm, *am, *r0, *r0s, *mjs, *d0, *d0s;
int vsz2;


/* Allocate global parallel variables, pair must have been allocated	*/
/* Calculates sizes.							*/

void create_globals()
 { fsz = boolsizeof(float:pair);
   vsz  = 3*fsz;  	/* vectors	  */
   cvsz = 10*fsz;	/* the conveyor	  */
   isz  = boolsizeof(int:pair);
   bsz  = boolsizeof(bool:pair);
   /* Float vectors[3] */
   x0  = palloc(pair, vsz);  x1 = palloc(pair, vsz);
   x0j = palloc(pair, vsz);
   v0  = palloc(pair, vsz);  v1 = palloc(pair, vsz);
   ac  = palloc(pair, vsz);  ax = palloc(pair, vsz);
   wv  = palloc(pair, vsz);
   /* xm0, xm1, acm, mmj must be consecutive */
   xm0 = palloc(pair, cvsz);
   xm1 = &xm0[3];  acm = &xm0[6];  mmj = &xm0[9];
   /* Floats */
   m  = palloc(pair, fsz);  mi = palloc(pair, fsz);  mj = palloc(pair, fsz);
   w  = palloc(pair, fsz);
   /* Ints and bools */
   i  = palloc(pair, isz);  j = palloc(pair, isz);  k = palloc(pair, isz);
   diag = palloc(pair, bsz);  act = palloc(pair, bsz);
 }

/*  Free all of the globals */

void free_globals()
 { /* Float vectors[3] */
   pfree(x0);  pfree(x1);  pfree(x0j);
   pfree(v0);  pfree(v1);
   pfree(ac);  pfree(ax);  pfree(wv);
   /* xm0, xm1, acm, mmj must be consecutive */
   pfree(xm0);  xm1 = acm = mmj = NULL;
   /* Floats */
   pfree(m);  pfree(mi);  pfree(mj);  pfree(w);
   /* Ints and bools */
   pfree(i);  pfree(j);  pfree(k);  pfree(diag);  pfree(act);
 }


/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      *i = pcoord(0);  *j = pcoord(1);  *k = pcoord(2);
      /* Identify active processors */
      *act = (*i != *j);  *diag = (*i == *j);
    }
 }


/* Print x-components of xx in nx by nx pipes (for testing). */

void print_x(char lbl[], float:current *xx, int nx)
 { int u, v, p;
   printf("\n%s\n", lbl); 
   for ( u = 0; u < nx; u++)
    { for ( v = 0; v < nx; v++)
       for ( p = 0; p < NS; p++)
        { if ( p%8 == 0 ) printf("\n");
          printf("%8.1e ", [u][v][p]*xx);
        }
      printf("\n");
    }
 }


/* Initialize masses, x0j. */

void initgreen()
 { int u;
   with (pair) everywhere
    { /* Spread scaled masses for particle j */
      *mj = [*j][*j][*k]*mi;
    }  /* End with (pair) */
 }


/* Control program for systolic acceleration calculation.	*/
/* Leaves Greenspan acceleration in Acg, which is calculated	*/
/* from X0, X1, gm of Particles.				*/

void acsmain(float:Particles Acg[3])
 { int u;
   /* Initialize the pair shape with suitable geometry */
   printf("Enter NM, NS for [NM][NM][NS] shape> ");
   scanf("%d%d", &NM, &NS);
   allocate_shape(&pair, 3, NM, NM, NS);
   SS = NS/2;
   create_globals();
   setup();
   with (pair) everywhere
     { int:pair p;
       /* Get positions and masses */
       p = *i*NS + *k;	/* position in Particles */
       *m = (float:pair) p;		/* For printing */
       /*	print_x("[0:1]][0:1][:]p[0]:", m, 2);  	*/
       *mi = [p]gm;	/* scaled mass */
       for ( u = 0;  u < 3;  u++ )
	{ x0[u] = [p]X0[u];  x1[u] = [p]X1[u]; }
       initgreen();
       /* print_x("[0:1]][0:1][:]x0[0]:", x0, 2);	/* For testing */
       /* print_x("[0:1]][0:1][:]x1[0]:", x1, 2);	/* For testing */
       /* print_x("[0:1]][0:1][:]*mi:", mi, 2);		/* For testing */
       /* print_x("[0:1]][0:1][:]*mj:", mj, 2);		/* For testing */
       CM_timer_clear(1);  CM_timer_start(1);
       spreadi(x0j, x0);
       accel();
       CM_timer_stop(1);
       printf("Direct acceleration calculation:\n");
       CM_timer_print(1);
       /*	print_x("[0:1]][0:1][:]ac[0]:", ac, 2);	*/
       /* Report result */
       where ( *j == 0 )
	for ( u = 0;  u < 3;  u++ )  [p]Acg[u] = ac[u];
     }  /* end with (pair) everywhere */
   free_globals();
   deallocate_shape(&pair);
 }
